package de.uni_passau.fim.seibt.gitwrapper.repo;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;

/**
 * A git {@link Repository}.
 */
public class Repository {

    private static final Logger LOG = Logger.getLogger(Repository.class.getCanonicalName());

    private static final String TYPE_COMMIT = "commit";

    private GitWrapper git;

    private String url;
    private File dir;

    private Map<String, Commit> commits;

    /**
     * Constructs a new {@link Repository}.
     *
     * @param git
     *         the {@link GitWrapper} to use
     * @param url
     *         the url the {@link Repository} was cloned from
     * @param dir
     *         the directory containing the {@link Repository}
     */
    Repository(GitWrapper git, String url, File dir) {
        this.git = git;
        this.url = url;

        try {
            this.dir = dir.getCanonicalFile();
        } catch (IOException e) {
            LOG.log(Level.WARNING, e, () -> "Exception computing the canonical file for " + dir + ". Using the absolute file instead.");
            this.dir = dir.getAbsoluteFile();
        }

        this.commits = new HashMap<>();
    }

    /**
     * Returns the merge commits in this repository.
     *
     * @return the merge commits
     */
    public List<Commit> getMergeCommits() {
        Optional<ExecRes> revList = git.exec(dir, "rev-list", "--all", "--merges");
        Function<ExecRes, List<Commit>> toCommitList = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to obtain the merge commits from %s.", this));
                return null;
            }

            String[] lines = res.output.split("[\\r?\\n]+");

            LOG.fine(() -> String.format("Found %d merge commits in %s.", lines.length, this));
            LOG.finer(() -> String.format("Merge commits are:%n%s", String.join(System.lineSeparator(), lines)));

            return Arrays.stream(lines).map(this::getCommitUnchecked).collect(Collectors.toList());
        };

        return revList.map(toCommitList).orElse(Collections.emptyList());
    }

    /**
     * Returns a {@link Commit} for the given ID. The caller must ensure that the ID is a full SHA1 hash of a
     * commit that exists in this repository.
     *
     * @param id the ID for the {@link Commit}
     * @return the {@link Commit}
     */
    Commit getCommitUnchecked(String id) {
        return commits.computeIfAbsent(id, theID -> new Commit(this, theID));
    }

    /**
     * Returns a {@link Commit} for the given ID. If the given ID does not designate a commit that exists in this
     * {@link Repository} an empty {@link Optional} will be returned. The ID will be resolved to a full SHA1 hash.
     *
     * @param id the ID of the commit
     * @return the {@link Commit} or an empty {@link Optional} if the ID is invalid or an exception occurs
     */
    public Optional<Commit> getCommit(String id) {

        if (commits.containsKey(id)) {
            return Optional.of(commits.get(id));
        }

        if (!isCommit(id)) {
            return Optional.empty();
        }

        return toHash(id).map(fullID -> commits.put(fullID, new Commit(this, fullID)));
    }

    /**
     * Determines whether the given object ID designates a commit.
     *
     * @param id
     *         the ID to check
     * @return true iff the ID designates a commit
     */
    private boolean isCommit(String id) {
        Optional<ExecRes> catFile = git.exec(dir, "cat-file", "-t", id);
        Function<ExecRes, Boolean> toBoolean = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to determine whether %s is a valid commit id.", id));
                return null;
            }

            boolean isCommit = res.output.startsWith(TYPE_COMMIT);

            if (isCommit) {
                LOG.finer(() -> String.format("%s is a commit.", id));
            }

            return isCommit;
        };

        return catFile.map(toBoolean).orElse(false);
    }

    /**
     * Resolves the given <code>id</code> to the full SHA1 hash.
     *
     * @param id
     *         the <code>id</code> to transform
     * @return the full SHA1 hash or an empty {@link Optional} if an exception occurs
     */
    private Optional<String> toHash(String id) {
        Optional<ExecRes> revParse = git.exec(dir, "rev-parse", id);
        Function<ExecRes, String> toHash = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to resolve %s to its unique SHA1 hash.", id));
                return null;
            }

            LOG.finer(() -> String.format("Resolved %s to %s.", id, res.output));

            return res.output;
        };

        return revParse.map(toHash);
    }

    /**
     * Returns the {@link GitWrapper} used by this {@link Repository}.
     *
     * @return the {@link GitWrapper}
     */
    GitWrapper getGit() {
        return git;
    }

    /**
     * Returns the URL this {@link Repository} was cloned from.
     *
     * @return the URL
     */
    public String getUrl() {
        return url;
    }

    /**
     * Returns the directory containing this {@link Repository}.
     *
     * @return the repository directory
     */
    public File getDir() {
        return dir;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Repository that = (Repository) o;
        return Objects.equals(url, that.url) && Objects.equals(dir, that.dir);
    }

    @Override
    public int hashCode() {
        return Objects.hash(url, dir);
    }

    @Override
    public String toString() {
        return String.format("Repository{url=%s, dir=%s}", url, dir);
    }
}
